import json
from typing import Union
import polars as pl

def write_schema(schema: Union[pl.DataFrame, pl.Schema], file: str):
    """Saves a Polars schema a JSON file
    
    Parameters
    ----------

    schema
        DataFrame or Schema Object.

    file 
        Save location and filename with JSON extension.

    Examples
    --------

    .. code-block:: python 

        import polars as pl
        import polars_extensions as plx
        data = pl.read_csv('datasets/employees.csv')
        plx.write_schema(data,'schema.json')
    """
    
    if isinstance(schema, pl.DataFrame):
        schema = schema.schema

    stringified_values = [str(value) for value in schema.dtypes()]
    schema_dict = dict(zip(schema.names(), stringified_values))

    with open(file, "w") as f:
        json.dump(schema_dict, f)
    return

def read_schema(file: str):
    """Opens a JSON Schema file and return a Polars Schema object
    
    Parameters
    ----------

    file 
        Save location and filename with JSON extension.

    Examples
    --------

    .. code-block:: python 

        import polars as pl
        import polars_extensions as plx
        schema = plx.read_schema('schema.json')
        schema

    
    """

    with open(file, "r") as f:
        schema = json.load(f)
    
    schema_dict = {}
    for k, v in schema.items():
        try:
            schema_dict[k] = getattr(pl, v)
        except AttributeError:
            raise ValueError(f"Invalid type {v} for column {k}")
    
    schema_object = pl.Schema(schema_dict)
    return schema_object


import polars as pl

@pl.api.register_dataframe_namespace("str_ext")
class StringExtensionNamespace:
    """String Extensions for the Polars Library"""

    def __init__(self, df: pl.DataFrame):
        self._df = df

    def f1_string_similarity(self, col_a: str, col_b: str) -> pl.DataFrame:
        """
        Calculates a similarity score between two columns of strings based on common characters,
        accounting for repeated characters.
        
        Parameters
        ----------
        col_a (str): The name of the first column to compare.
        col_b (str): The name of the second column to compare.
        
        Returns
        -------
        DataFrame

        Examples
        --------

        .. code-block:: python

            import polars_extensions as plx
            import polars as pl 

            data = pl.read_csv('datasets/string_sim.csv')
            data.str_ext.f1_string_similarity('a','c')

        .. code-block:: text

            shape: (13, 3)
            ┌──────────────────────────┬───────────────────────────┬──────────┐
            │ a                        ┆ c                         ┆ f1_score │
            │ ---                      ┆ ---                       ┆ ---      │
            │ str                      ┆ str                       ┆ f64      │
            ╞══════════════════════════╪═══════════════════════════╪══════════╡
            │ apple                    ┆ appl                      ┆ 0.888889 │
            │ banana                   ┆ BANANA                    ┆ 1.0      │
            │ cherry                   ┆ cherr                     ┆ 0.909091 │
            │ date                     ┆ etad                      ┆ 1.0      │
            │ elderberry               ┆ elderberrys               ┆ 0.952381 │
            │ …                        ┆ …                         ┆ …        │
            │ kiwi                     ┆ KIW                       ┆ 0.857143 │
            │ lemon                    ┆ lemons                    ┆ 0.909091 │
            │ mangoes are Tangy        ┆ mango are Tangy           ┆ 0.9375   │
            │ it was the best of times ┆ it was the worst of times ┆ 0.897959 │
            │ of times it was the best ┆ it was the worst of times ┆ 0.897959 │
            └──────────────────────────┴───────────────────────────┴──────────┘
        

        """

        def similarity(row_str_a: str, row_str_b: str) -> float:
            # Normalize both strings (case-insensitive comparison)
            row_str_a = row_str_a.lower()
            row_str_b = row_str_b.lower()

            # If strings are identical, return a score of 1.0
            if row_str_a == row_str_b:
                return 1.0

            list1 = list(row_str_a)
            list2 = list(row_str_b)

            list2_copy = list2[:]
            intersection = []

            # Account for repeated characters by checking all occurrences
            for char in list1:
                if char in list2_copy:
                    intersection.append(char)
                    list2_copy.remove(char)
            
            common_chars = len(intersection)
            total_chars = len(list1) + len(list2)
            return (2 * common_chars) / total_chars if total_chars > 0 else 0.0

        # Apply the similarity function row-by-row
        similarity_scores = [
            similarity(row_a, row_b) for row_a, row_b in zip(self._df[col_a], self._df[col_b])
        ]

        # Add the similarity scores as a new column to the DataFrame
        self._df = self._df.with_columns(
            pl.Series("f1_score", similarity_scores)
        )

        return self._df


from livereload import Server, shell
from source.conf import html_static_path, templates_path

# -------------------------------------------------------------------------
# To use, just execute `python run_live_docs_server.py` in a terminal
# and a local server will run the docs in your browser, automatically
# refreshing/reloading the pages you're working on as they are modified.
# Extremely helpful to see the real output before it gets uploaded, and
# a much smoother experience than constantly running `make html` yourself.
# -------------------------------------------------------------------------

if __name__ == "__main__":
    # establish a local docs server
    svr = Server()

    # command to rebuild the docs
    refresh_docs = shell("make html")

    # watch for source file changes and trigger rebuild/refresh
    svr.watch("*.rst", refresh_docs, delay=1)
    svr.watch("*.md", refresh_docs, delay=1)
    svr.watch("source/reference/*", refresh_docs, delay=1)
    for path in html_static_path + templates_path:
        svr.watch(f"source/{path}/*", refresh_docs, delay=1)

    # path from which to serve the docs
    svr.serve(root="build/html")


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from __future__ import annotations

import inspect
import os
import re
import sys
import warnings
from pathlib import Path
from typing import Any


import logging

# Configure the logger
logging.basicConfig(
    filename='docs.log',  # Log file name
    level=logging.DEBUG,  # Log level
    format='%(asctime)s - %(levelname)s - %(message)s'  # Log message format
)


# import sphinx_autosummary_accessors

# -- Path setup --------------------------------------------------------------

sys.path.insert(0, str(Path("../..").resolve()))


# -- Project information -----------------------------------------------------

project = 'Polars Extensions'
copyright = '2025, Jayden Rasband'
author = 'Jayden Rasband'
release = '0.8.0'

# -- General configuration ---------------------------------------------------

extensions = [
    # Sphinx extensions
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.githubpages",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.mathjax",
    # Third-party extensions
    "autodocsumm",
    "numpydoc",
    # "sphinx_autosummary_accessors",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_favicon",
    "sphinx_reredirects",
    "sphinx_toolbox.more_autodoc.overloads",
    "recommonmark"
]

# Render docstring text in `single backticks` as code.
default_role = "code"

maximum_signature_line_length = 88

# Below setting is used by
# sphinx-autosummary-accessors - build docs for namespace accessors like `Series.str`
# https://sphinx-autosummary-accessors.readthedocs.io/en/stable/
templates_path = ["_templates", 
                #   sphinx_autosummary_accessors.templates_path
                  ]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["Thumbs.db", ".DS_Store"]

# Hide overload type signatures
# sphinx_toolbox - Box of handy tools for Sphinx
# https://sphinx-toolbox.readthedocs.io/en/latest/
overloads_location = ["bottom"]
language = 'python'


# -- Options for HTML output -------------------------------------------------


html_theme = 'alabaster'
html_static_path = ['_static']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]  # relative to html_static_path
html_show_sourcelink = False

# key site root paths
static_assets_root = ""
github_root = "https://github.com/jrasband-dev/polars-extensions"
web_root = "https://jrasband-dev.github.io/polars-extensions/"
asset_root = "https://raw.githubusercontent.com/jrasband-dev/polars-extensions/main"

# Specify version for version switcher dropdown menu
git_ref = os.environ.get("METRIC_FORGE_VERSION", "main")
version_match = re.fullmatch(r"py-(\d+)\.\d+\.\d+.*", git_ref)
switcher_version = version_match.group(1) if version_match is not None else "dev"

# html_js_files = [
#     (
#         "https://plausible.io/js/script.js",
#         {"data-domain": "docs.pola.rs,combined.pola.rs", "defer": "defer"},
#     ),
# ]

html_theme_options = {
    # "external_links": [
    #     {
    #         "name": "User guide",
    #         "url": f"{web_root}/",
    #     },
    # ],
    "icon_links": [
        {
            "name": "GitHub",
            "url": github_root,
            "icon": "fa-brands fa-github",
        },
        {
            "name": "Pypi",
            "url": "https://pypi.org/project/polars-extensions/",
            "icon": "fa-brands fa-python",
        },
        {
            "name": "LinkedIn",
            "url": "https://www.linkedin.com/in/jayden-rasband-303449133/",
            "icon": "fa-brands fa-linkedin",
        },
        {
            "name": "YouTube",
            "url": "https://www.youtube.com/@polarsCodeAcademy",
            "icon": "fa-brands fa-youtube",
        },
    ],
    # "logo": {
    #     "image_light": f"{asset_root}/static/metric-forge-black.png",
    #     "image_dark": f"{asset_root}/static/metric-forge-white.png",
    # },
    # "switcher": {
    #     "json_url": f"{web_root}/api/python/dev/_static/version_switcher.json",
    #     "version_match": switcher_version,
    # },
    "show_version_warning_banner": False,
    "navbar_end": ["theme-switcher", "version-switcher", "navbar-icon-links"],
    "check_switcher": False,
}

# sphinx-favicon - Add support for custom favicons
# https://github.com/tcmetzger/sphinx-favicon

# favicons = [
#     {
#         "rel": "icon",
#         "sizes": "32x32",
#         "href": f"{static_assets_root}/icons/favicon-32x32.png",
#     },
#     {
#         "rel": "apple-touch-icon",
#         "sizes": "180x180",
#         "href": f"{static_assets_root}/icons/touchicon-180x180.png",
#     },
# ]


# sphinx-ext-linkcode - Add external links to source code
# https://www.sphinx-doc.org/en/master/usage/extensions/linkcode.html
def linkcode_resolve(domain: str, info: dict[str, Any]) -> str | None:
    """
    Determine the URL corresponding to Python object.

    Based on pandas equivalent:
    https://github.com/pandas-dev/pandas/blob/main/doc/source/conf.py#L629-L686
    """
    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            with warnings.catch_warnings():
                # Accessing deprecated objects will generate noisy warnings
                warnings.simplefilter("ignore", FutureWarning)
                obj = getattr(obj, part)
        except AttributeError:
            return None

    try:
        fn = inspect.getsourcefile(inspect.unwrap(obj))
    except TypeError:
        try:  # property
            fn = inspect.getsourcefile(inspect.unwrap(obj.fget))
        except (AttributeError, TypeError):
            fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except TypeError:
        try:  # property
            source, lineno = inspect.getsourcelines(obj.fget)
        except (AttributeError, TypeError):
            lineno = None
    except OSError:
        lineno = None

    linespec = f"#L{lineno}-L{lineno + len(source) - 1}" if lineno else ""

    conf_dir_path = Path(__file__).absolute().parent
    project_root = (conf_dir_path.parent.parent / "polars-extensions").absolute()

    fn = os.path.relpath(fn, start=project_root)
    return f"{github_root}/blob/{git_ref}/polars-extensions/{fn}{linespec}"


def _minify_classpaths(s: str) -> str:
    logging.debug(f'classpath: {s}')
    return re.sub(
        pattern=r"metric_forge\.ecommerce\.([^.]+\.[^.]+)",  
        repl=r"\1",  # Replace with just the captured part
        string=s,
    )



def process_signature(  # noqa: D103
    app: object,
    what: object,
    name: object,
    obj: object,
    opts: object,
    sig: str,
    ret: str,
) -> tuple[str, str]:
    logging.debug(f'process_sig: {sig}')
    logging.debug(f'return_sig: {ret}')
    return (
        _minify_classpaths(sig) if sig else sig,
        _minify_classpaths(ret) if ret else ret,
    )


def setup(app: Any) -> None:  # noqa: D103
    app.connect("autodoc-process-signature", process_signature)


import polars as pl



@pl.api.register_expr_namespace("num_ext") 
class NumericExtensionNamespace: 
    def __init__(self, expr: pl.Expr):
        self._expr = expr

    # Roman numeral mappings
    _roman_map = [
        ("M", 1000), ("CM", 900), ("D", 500), ("CD", 400),
        ("C", 100), ("XC", 90), ("L", 50), ("XL", 40),
        ("X", 10), ("IX", 9), ("V", 5), ("IV", 4), ("I", 1)
    ]

    def to_roman(self) -> pl.Expr:
        """
        Convert an integer to Roman numerals.

        Examples
        --------
        .. code-block:: python
        
            import polars as pl
            import polars_extensions as plx
            df = pl.DataFrame({"numbers": [1, 2, 309, 4, 5]})
            result = df.with_columns(
                pl.col('numbers').num_ext.to_roman().alias("Roman")
            )

            result

        .. code-block:: text 

            shape: (5, 2)
            ┌─────────┬───────┐
            │ numbers ┆ Roman │
            │ ---     ┆ ---   │
            │ i64     ┆ str   │
            ╞═════════╪═══════╡
            │ 1       ┆ I     │
            │ 2       ┆ II    │
            │ 309     ┆ CCCIX │
            │ 4       ┆ IV    │
            │ 5       ┆ V     │
            └─────────┴───────┘
        """
        def convert_to_roman(value: int) -> str:
            if not (0 < value < 4000):
                raise ValueError("Number out of range (1-3999)")

            result = []
            for roman, num in self._roman_map:
                while value >= num:
                    result.append(roman)
                    value -= num
            return ''.join(result)

        # Use map_elements for element-wise operation
        return self._expr.map_elements(convert_to_roman,return_dtype=pl.String)

    def from_roman(self) -> pl.Expr:
        """
        Convert Roman numerals to integers.

        Examples
        --------
        .. code-block:: python
        
            import polars_extensions as plx

            df = pl.DataFrame({"Roman": ['I', 'II', 'III', 'CCCIX', 'V']})
            result = df.with_columns(
                pl.col('Roman').num_ext.from_roman().alias("Decoded")
            )

            result

        .. code-block:: text 

            shape: (5, 2)
            ┌───────┬─────────┐
            │ Roman ┆ Decoded │
            │ ---   ┆ ---     │
            │ str   ┆ i64     │
            ╞═══════╪═════════╡
            │ I     ┆ 1       │
            │ II    ┆ 2       │
            │ III   ┆ 3       │
            │ CCCIX ┆ 309     │
            │ V     ┆ 5       │
            └───────┴─────────┘
        """
        roman_to_value = {roman: value for roman, value in self._roman_map}

        def convert_from_roman(roman: str) -> int:
            i = 0
            total = 0
            while i < len(roman):
                # Check for two-character numeral first
                if i + 1 < len(roman) and roman[i:i+2] in roman_to_value:
                    total += roman_to_value[roman[i:i+2]]
                    i += 2
                else:
                    total += roman_to_value[roman[i]]
                    i += 1
            return total

        # Use map_elements for element-wise operation
        return self._expr.map_elements(convert_from_roman,return_dtype=pl.Int64)
    
    def word_to_number(self) -> pl.Expr:
        """Convert Natural Language to Numbers

        Examples
        --------

        .. code-block:: python

            import polars as pl
            import polars_extensions as plx

            df = pl.DataFrame({"numbers": ['6', 'two', 'three hundred and nine', '5', '4']})
            df.with_columns(
                pl.col('numbers').num_ext.word_to_number().alias("Actual Numbers")
            )

        .. code-block:: text 

            shape: (5, 2)
            ┌────────────────────────┬────────────────┐
            │ numbers                ┆ Actual Numbers │
            │ ---                    ┆ ---            │
            │ str                    ┆ i64            │
            ╞════════════════════════╪════════════════╡
            │ 6                      ┆ 6              │
            │ two                    ┆ 2              │
            │ three hundred and nine ┆ 309            │
            │ 5                      ┆ 5              │
            │ 4                      ┆ 4              │
            └────────────────────────┴────────────────┘
            
        
        """
        from word2number import w2n
        
        return self._expr.map_elements(lambda x: w2n.word_to_num(x) if isinstance(x, str) else x, return_dtype=pl.Int64)
